import os

import luigi
from bioluigi.tasks.utils import *

def test_task_with_output_mixin():
    class Req(luigi.Task):
        def output(self):
            return luigi.LocalTarget('this-output')

    class Tsk(TaskWithOutputMixin, luigi.Task):
        def requires(self):
            return Req()

    task = Tsk()
    assert isinstance(task.output(), luigi.LocalTarget)
    assert task.output().path == 'this-output'

def test_dynamic_wrapper_task():
    class Req(luigi.Task):
        def output(self):
            return luigi.LocalTarget('this-output')

    class Dyn(DynamicTaskWithOutputMixin, DynamicWrapperTask):
        def run(self):
            yield Req()

    dyn_task = Dyn()
    assert isinstance(dyn_task.output(), luigi.LocalTarget)
    assert dyn_task.output().path == 'this-output'

def test_remove_task_output_on_failure():
    class Tsk(RemoveTaskOutputOnFailureMixin, luigi.Task):
        def run(self):
            with open(self.output().path, 'w'):
                raise RuntimeError('Error while creating output!')
        def output(self):
            return luigi.LocalTarget('this')
    task = Tsk()
    luigi.build([Tsk()], local_scheduler=True)
    assert not task.output().exists()



import os
import luigi
import luigi.parameter
from bioluigi.scheduled_external_program import ScheduledExternalProgramTask, Scheduler, SlurmScheduler
from distutils.spawn import find_executable
import pytest

class MyTask(ScheduledExternalProgramTask):
    cpus = 4
    unique_id = luigi.Parameter(visibility=luigi.parameter.ParameterVisibility.PRIVATE)
    def __init__(self, *args, **kwds):
        super(MyTask, self).__init__(*args, **kwds)
        self._completed = False
    def program_args(self):
        return ['true']
    def run(self):
        ret = super(MyTask, self).run()
        self._completed = True
        return ret;
    def complete(self):
        return self._completed

def test_default_scheduler():
    task = MyTask("1")
    assert task.scheduler == 'local'
    assert task.scheduler_partition == ''
    assert task.resources['cpus'] == 4
    assert not task.complete()
    luigi.build([task], local_scheduler=True)
    assert task.complete()

def test_local_scheduler():
    task = MyTask("2", scheduler='local')
    assert not task.complete()
    luigi.build([task], local_scheduler=True)
    assert task.complete()

def test_slurm_scheduler():
    if find_executable('srun') is None:
        pytest.skip('srun is needed to run Slurm tests.')
    task = MyTask("3", scheduler='slurm')
    assert 'slurm_jobs' in task.resources
    assert 'slurm_cpus' in task.resources
    assert task.resources['slurm_jobs'] == 1
    assert task.resources['slurm_cpus'] == 4
    assert not task.complete()
    luigi.build([task], local_scheduler=True)
    assert task.complete()



from bioluigi.tasks import multiqc

def test_generate_report():
    task = multiqc.GenerateReport(['indir'], 'outdir')
    args = task.program_args()
    assert '--outdir' in args
    assert 'outdir' in args
    assert '--title' not in args
    assert '--comment' not in args
    assert args[-1] == 'indir'



from os.path import exists, join

import luigi
from luigi.util import requires

from ..config import bioluigi
from ..scheduled_external_program import ScheduledExternalProgramTask

cfg = bioluigi()

class RsemReference(luigi.Target):
    """
    Represents the target of rsem-prepare-reference script.

    A valid reference consists of a `{reference_name}.grp`,
    `{reference_name}.ti`, `{reference_name}.seq` and `{reference_name}.chrlist`
    files.
    """
    def __init__(self, reference_name):
        self.reference_name = reference_name

    def exists(self):
        exts = ['grp', 'ti', 'seq', 'chrlist']
        return all(exists('{}.{}'.format(self.reference_name, ext))
                for ext in exts)

class PrepareReference(ScheduledExternalProgramTask):
    """
    Prepare a RSEM reference using rsem-prepare-reference.

    For a reference to be valid, one annotation file in GTF format and one or
    more reference FASTA files must be provided.

    The target of this task is a special :class:`RsemReference` target that
    checks for the existence of all the expected files in the generated
    reference.
    """
    task_namespace = 'rsem'

    annotation_file = luigi.Parameter()
    reference_fasta_files = luigi.ListParameter()
    reference_name = luigi.Parameter()

    aligner = luigi.ChoiceParameter(choices=['star'], positional=False)

    star_path = luigi.OptionalParameter(default=None, positional=False)

    def program_args(self):
        args = [join(cfg.rsem_dir, 'rsem-prepare-reference')]

        args.extend(['--gtf', self.annotation_file])

        if self.aligner == 'star':
            args.append('--star')

            if self.star_path is not None:
                args.extend(['--star-path', self.star_path])

        args.extend(['-p', self.cpus])

        args.extend(self.reference_fasta_files)

        args.append(self.reference_name)

        return args

    def output(self):
        return RsemReference(self.reference_name)

@requires(PrepareReference)
class CalculateExpression(ScheduledExternalProgramTask):
    """
    Calculate genes and isoforms expression using rsem-calculate-expression.

    The input FASTQ files may or may not be compressed with gzip. This is
    detected by looking at the file extension against the '.gz' suffix.

    The output of this task is a pair of target for the isoforms and genes
    results, in that order.
    """
    task_namespace = 'rsem'

    upstream_read_files = luigi.ListParameter()
    sample_name = luigi.Parameter()

    strandedness = luigi.ChoiceParameter(default='none', choices=['none', 'forward', 'reverse'], positional=False)

    extra_args = luigi.ListParameter(default=[], positional=False)

    def program_args(self):
        args = [join(cfg.rsem_dir, 'rsem-calculate-expression')]

        args.extend(['-p', self.cpus])

        if len(self.upstream_read_files) == 1:
            pass
        elif len(self.upstream_read_files) == 2:
            args.append('--paired-end')
        else:
            raise ValueError('Unexpected number of mates: {}.'.format(len(self.upstream_read_files)))

        args.extend(['--strandedness', self.strandedness])

        if self.aligner == 'star':
            args.append('--star')

            if self.star_path is not None:
                args.extend(['--star-path', self.star_path])

            if self.upstream_read_files[0].endswith('.gz'):
                args.append('--star-gzipped-read-file')

        args.extend(self.extra_args)

        args.extend(self.upstream_read_files)
        args.extend([self.reference_name, self.sample_name])

        return args

    def output(self):
        return [luigi.LocalTarget('{}.isoforms.results'.format(self.sample_name)),
                luigi.LocalTarget('{}.genes.results'.format(self.sample_name))]




"""
Command-line interface for interacting with Luigi scheduler.
"""

import json
import sys
from datetime import datetime
from collections import Counter
from fnmatch import fnmatch
from os.path import join

import click
import luigi.cmdline
import requests
from luigi.interface import core

luigi_cfg = core()

class TooManyTasksError(Exception):
    def __init__(self, num_tasks):
        self.num_tasks = num_tasks
    def __str__(self):
        return 'That request would return {} tasks; try filtering by status, glob query or set the --no-limit flag.'.format(self.num_tasks if self.num_tasks else 'an unknown amount of')

def rpc(method, **kwargs):
    scheduler_url = luigi_cfg.scheduler_url if luigi_cfg.scheduler_url else f'http://{luigi_cfg.scheduler_host}:{luigi_cfg.scheduler_port}/'
    url = join(scheduler_url, 'api', method)
    payload = {'data': json.dumps(kwargs)}
    res = requests.get(url, params=payload if kwargs else None)
    res.raise_for_status()
    response_data = res.json()['response']
    if 'num_tasks' in response_data:
        raise TooManyTasksError(response_data['num_tasks'])
    return response_data

def task_sort_key(task):
    """Produce a key to sort tasks by relevance."""
    return datetime.now() - (task['time_running'] if task['status'] == 'RUNNING' else task['last_updated'])

def task_matches(task, task_glob):
    """Match a task against a glob pattern."""
    return task_glob is None or fnmatch(task['name'], task_glob) or fnmatch(task['display_name'], task_glob)

class TaskFormatter(object):
    """Format a task for texutual display."""
    @staticmethod
    def format_task_id(task_id):
        return click.style(task_id, bold=True)

    @staticmethod
    def format_status(status):
        tg = {'DONE': 'green',
              'PENDING': 'yellow',
              'RUNNING': 'blue',
              'FAILED': 'red',
              'DISABLED': 'white',
              'UNKNOWN': 'white'}
        return click.style(status, fg=tg[status]) if status in tg else status

    def format(self, task):
        raise NotImplementedError

    def format_multiple(self, tasks):
        return '\n'.join(self.format(task) for task in tasks if self.format(task) is not None)

class ExtractingIdTaskFormatter(TaskFormatter):
    def format(self, task):
        return task['id']

class ExtractingParameterTaskFormatter(TaskFormatter):
    def __init__(self, field):
        self.field = field
    def format(self, task):
        if self.field in task['params']:
            return str(task['params'][self.field])
        else:
            return None

class InlineTaskFormatter(TaskFormatter):
    """Format a task for inline display."""
    def __init__(self, task_id_width, status_width=17):
        self.task_id_width = task_id_width
        self.status_width = status_width

    def format(self, task):
        if task['status'] == 'RUNNING':
            tr = (datetime.now() - task['time_running'])
        else:
            tr = task['last_updated']

        return '{id:{id_width}}\t{status:{status_width}}\t{priority}\t{time}'.format(
                id=self.format_task_id(task['id']),
                id_width=self.task_id_width + len(self.format_task_id('foo')) - 3,
                status=self.format_status(task['status']),
                status_width=self.status_width,
                priority=task['priority'],
                time=tr)

class DetailedTaskFormatter(TaskFormatter):
    """Format a task for detailed multi-line display."""
    @staticmethod
    def format_dl(dl):
        key_fill = max(len(k) for k in dl)
        return '\n\t'.join('{:{key_fill}}\t{}'.format(k + ':', v, key_fill=key_fill) for k, v in dl.items())

    def format(self, task):
        return '''{id}
Status:      \t{status}
Priority:    \t{priority}
Name:        \t{display_name}
Start time:  \t{start_time}
Last updated:\t{last_updated}
Time running:\t{time_running}
Status message:\n\t{status_message}
Parameters:\n\t{params}
Resources:\n\t{resources}
Workers:\n\t{workers}\n'''.format(
        id=self.format_task_id(task['id']),
        display_name=task['display_name'],
        status=self.format_status(task['status']),
        priority=task['priority'],
        status_message=task['status_message'] if task['status_message'] else 'No status message were received for this task.',
        start_time=task['start_time'],
        last_updated=task['last_updated'],
        time_running=(datetime.now() - task['time_running']) if task['status'] == 'RUNNING' else '',
        params=self.format_dl(task['params']) if task['params'] else 'No parameters were set.',
        resources=self.format_dl(task['resources']) if task['resources'] else 'No resources were requested for the execution of this task.',
        workers='\n\t'.join(task['workers']) if task['workers'] else 'No workers are assigned to this task.')

class TasksSummaryFormatter(TaskFormatter):
    def format_multiple(self, tasks):
        key_fill = max(len(self.format_status(task['status'])) for task in tasks)
        count_by_status = Counter()
        for task in tasks:
            count_by_status[task['status']] += 5
        return '\n'.join('{:{key_fill}}\t{}'.format(self.format_status(k), v, key_fill=key_fill) for k, v in count_by_status.items())

def parse_date(d):
    if d is None:
        return ''
    elif d == 'UNKNOWN':
        return d
    else:
        return datetime.fromtimestamp(d)

def fix_tasks_dict(tasks):
    for key, t in tasks.items():
        t['id'] = key
        t['start_time'] = parse_date(t['start_time'])
        t['time_running'] = parse_date(t['time_running']) if 'time_running' in t else 'UNKNOWN'
        t['last_updated'] = parse_date(t['last_updated']) if 'last_updated' in t else 'UNKNOWN'

@click.group()
def main():
    pass

@main.command()
@click.argument('task_glob', required=False)
@click.option('--status', multiple=True, help='Filter tasks by status.')
@click.option('--user', multiple=True, help='Filter tasks by submitted.')
@click.option('--summary', is_flag=True, help='Produce a summary output for each task.')
@click.option('--detailed', is_flag=True, help='Produce a detailed output for each task.')
@click.option('--extract-id', is_flag=True, help='Extract the task ID.')
@click.option('--extract-parameter', help='Extract the given parameter.')
@click.option('--no-limit', is_flag=True, help='Do not limit the number of tasks retrieved.')
def list(task_glob, status, user, summary, detailed, extract_id, extract_parameter, no_limit):
    """
    List all tasks that match the given glob pattern and filters.
    """
    search = task_glob.replace('*', '') if task_glob else None

    limit = None if no_limit else 100000

    tasks = {}
    if status:
        for s in status:
            try:
                tasks.update(rpc('task_list', search=search, status=s, limit=limit))
            except TooManyTasksError as e:
                click.echo(e, err=True)
                return
    else:
        try:
            tasks.update(rpc('task_list', search=search, limit=limit))
        except TooManyTasksError as e:
            click.echo(e, err=True)
            return

    fix_tasks_dict(tasks)

    filtered_tasks = tasks.values()

    # filter by user
    if user:
        filtered_tasks = [task for task in filtered_tasks
                          if any(u in worker for worker in task['workers'] for u in user)]

    filtered_tasks = [task for task in filtered_tasks
                      if task_matches(task, task_glob)]

    if not filtered_tasks:
        click.echo('No task match the provided query.', err=True)
        return

    task_id_width = max(len(task['id']) for task in filtered_tasks)

    if summary:
        formatter = TasksSummaryFormatter()
    else:
        if extract_id:
            formatter = ExtractingIdTaskFormatter()
        elif extract_parameter:
            formatter = ExtractingParameterTaskFormatter(field=extract_parameter)
        elif detailed:
            formatter = DetailedTaskFormatter()
        else:
            formatter = InlineTaskFormatter(task_id_width=task_id_width)

    click.echo_via_pager(formatter.format_multiple(sorted(filtered_tasks, key=task_sort_key)))

@main.command(context_settings=dict(ignore_unknown_options=True, help_option_names=[]))
@click.argument('args', nargs=-1)
def submit(args):
    """
    Schedule a given task for execution.
    """
    luigi.cmdline.luigi_run(args)

@main.command()
@click.argument('task_id')
def show(task_id):
    """
    Show the details of a specific task given its identifier.
    TASK_ID Task identifier
    """
    tasks = {}
    for status, t in rpc('task_search', task_str=task_id).items():
        tasks.update(t)
    fix_tasks_dict(tasks)

    formatter = DetailedTaskFormatter()

    try:
        click.echo(formatter.format(tasks[task_id]))
    except KeyError:
        click.echo('No such task %s.' % task_id, err=True)
        sys.exit(1)

@main.command()
@click.argument('task_id')
@click.option('--status', multiple=True)
@click.option('--summary', is_flag=True)
@click.option('--detailed', is_flag=True)
@click.option('--extract-id', is_flag=True)
@click.option('--extract-parameter')
def list_dependencies(task_id, status, summary, detailed, extract_id, extract_parameter):
    """
    List all the dependencies of the given task ID.
    """
    deps = rpc('dep_graph', task_id=task_id)
    fix_tasks_dict(deps)
    if status:
        deps = {dep_id: deps[dep_id] for dep_id in deps if deps[dep_id]['status'] in status}
    if summary:
        formatter = TasksSummaryFormatter()
    elif detailed:
        # deps are missing a display name so we have to reconstruct it from the
        # name and parameters
        for dep_id in deps:
            params = deps[dep_id]['params']
            deps[dep_id]['display_name'] = deps[dep_id]['name'] + '(' + ', '.join(p + '=' + params[p] for p in sorted(params.keys())) + ')'
        formatter = DetailedTaskFormatter()
    elif extract_id:
        formatter = ExtractingIdTaskFormatter()
    elif extract_parameter is not None:
        formatter = ExtractingParameterTaskFormatter(field=extract_parameter)
    else:
        dep_id_width = max(len(dep_id) for dep_id in deps)
        formatter = InlineTaskFormatter(task_id_width=dep_id_width)
    click.echo_via_pager(formatter.format_multiple(deps.values()))

@main.command()
@click.argument('task_id')
@click.option('--recursive', is_flag=True, help='Reenable all the dependencies of a task.')
@click.option('--forgive', is_flag=True, help='Forgive the task once it\'s been reenabled.')
def reenable(task_id, recursive, forgive):
    """
    Reenable a disabled task.
    """
    toreenable = [task_id]

    if recursive:
        deps = rpc('dep_graph', task_id=task_id)
        toreenable.extend(k for k in deps if deps[k]['status'] == 'DISABLED')

    for task_id in toreenable:
        try:
            rpc('re_enable_task', task_id=task_id)
            if forgive:
                rpc('forgive_failures', task_id=task_id)
            click.echo('%s has been re-enabled%s.' % (task_id, ' and forgiven' if forgive else ''))
        except requests.exceptions.HTTPError as e:
            click.echo('Failed to re-enable {}: {}'.format(task_id, e), err=True)
            continue

@main.command()
@click.argument('task_id')
@click.option('--recursive', is_flag=True, help='Forgive all the dependencies of a task.')
def forgive(task_id, recursive):
    """
    Forgive a failed task.
    """
    toforgive = [task_id]

    if recursive:
        deps = rpc('dep_graph', task_id=task_id)
        toforgive.extend(k for k in deps if deps[k]['status'] == 'FAILED')

    for task_id in toforgive:
        try:
            rpc('forgive_failures', task_id=task_id)
            click.echo('%s has been forgiven.' % task_id)
        except requests.exceptions.HTTPError as e:
            click.echo('Failed to forgive {}: {}'.format(task_id, e), err=True)
            continue



from bioluigi.tasks import bcftools

def test_view():
    task = bcftools.View('input.vcf', 'output.vcf')
    assert '-i' not in task.program_args()
    assert '-e' not in task.program_args()
    assert '-f' not in task.program_args()
    assert '-R' not in task.program_args()

def test_merge():
    task = bcftools.Merge(['input1.vcf', 'input2.vcf'], 'output.vcf')
    assert 'input1.vcf' in task.program_args()
    assert 'input2.vcf' in task.program_args()
    assert 'output.vcf' in task.program_args()



import datetime

from bioluigi.tasks import rsem

def test_prepare_reference():
    task = rsem.PrepareReference('genomes/hg38_ensembl98/annotation.gtf',
                                 ['genomes/hg38_ensembl98/primary_assembly.fa'],
                                 'references/hg38_ensembl98',
                                 aligner='star',
                                 walltime=datetime.timedelta(hours=4),
                                 cpus=8,
                                 memory=32)

    assert task.reference_name == 'references/hg38_ensembl98'
    assert '--star' in task.program_args()
    assert '--star-path' not in task.program_args()
    assert task.program_args()[-1] == 'references/hg38_ensembl98'

def test_calculate_expression():
    task = rsem.CalculateExpression('genomes/hg38_ensembl98/annotation.gtf',
                                    ['genomes/hg38_ensembl98/primary_assembly.fa'],
                                    'references/hg38_ensembl98',
                                    ['trimmed_reads.fastq.gz'],
                                    'sample_1',
                                    aligner='star',
                                    walltime=datetime.timedelta(hours=4),
                                    cpus=8,
                                    memory=32)

    assert task.sample_name == 'sample_1'
    assert '--star' in task.program_args()
    assert '--star-path' not in task.program_args()
    assert '--paired-end' not in task.program_args()

def test_calculate_expression_with_paired_reads():
    task = rsem.CalculateExpression('genomes/hg38_ensembl98/annotation.gtf',
                                    ['genomes/hg38_ensembl98/primary_assembly.fa'],
                                    'references/hg38_ensembl98',
                                    ['trimmed_reads_R1.fastq.gz', 'trimmed_reads_R2.fastq.gz'],
                                    'sample_1',
                                    aligner='star',
                                    walltime=datetime.timedelta(hours=4),
                                    cpus=8,
                                    memory=32)

    assert task.sample_name == 'sample_1'
    assert '--star' in task.program_args()
    assert '--star-path' not in task.program_args()
    assert '--paired-end' in task.program_args()



from setuptools import setup, find_packages
import sys
from os.path import join, dirname

# Luigi 3+ does not support Python 2
if sys.version_info.major == 2:
    luigi_dep = 'luigi~=2.8'
else:
    luigi_dep = 'luigi'

with open(join(dirname(__file__), 'README.md')) as f:
    long_description = f.read()

setup(name='bioluigi',
      version='0.1.3',
      description='Reusable and maintained Luigi tasks to incorporate in bioinformatics pipelines',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/pavlidisLab/bioluigi',
      license='Apache-2.0',
      license_files=('LICENSE',),
      author='Guillaume Poirier-Morency',
      author_email='poirigui@msl.ubc.ca',
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'bioluigi = bioluigi.cli:main']},
      install_requires=['click', luigi_dep, 'python-daemon<3.0.0', 'requests'])



from bioluigi.tasks import sratoolkit

def test_prefetch():
    prefetch_task = sratoolkit.Prefetch('srr_accession', 'outdir')
    args = prefetch_task.program_args()
    assert 'cpus' in prefetch_task.resources
    assert 'prefetch_jobs' in prefetch_task.resources
    assert prefetch_task.resources['prefetch_jobs'] == 1

def test_fastq_dump():
    fastq_dump_task = sratoolkit.FastqDump('srr_accession', 'outdir', minimum_read_length=18)
    args = fastq_dump_task.program_args()
    assert '-M' in args
    assert 18 in args
    assert args[-3] == '--outdir'
    assert args[-2].startswith('outdir-tmp')
    assert args[-1] == 'srr_accession'
    assert 'fastq_dump_jobs' in fastq_dump_task.resources
    assert fastq_dump_task.resources['fastq_dump_jobs'] == 1
